#!/bin/bash

set -e

# -k means keep the kernel kompat kode
# -kk means run kernel_compat.sh at export time
# -o means keep out-of-tree (not-upstream) code
# --suffix=<identifier> adds a suffix to the EFX_DRIVER_VERSION define
SOURCE_SED_COMMANDS=''
MAKEFILE_SED_COMMANDS=''
UNIFDEF_DEFINES=''
SUBDIR=drivers/base
KCOMPAT=
AUTOCOMPAT=
VERSION_SUFFIX=
NOT_UPSTREAM=

test -d $KPATH/drivers/base && SUBDIR=drivers/base || true

for i in "$@"
do
case $i in
	-k)
		KCOMPAT=1
		;;
	-kk)
		KCOMPAT=1
		AUTOCOMPAT=apply
		;;
	-o)
		AUTOCOMPAT=
		NOT_UPSTREAM=1
		;;
	-s)	shift
		SUBDIR=$1
		;;
	--suffix=*)
		VERSION_SUFFIX="${i#*=}"
		;;
	*)
		# End of arguments
		break
		;;
esac
shift
done

if [ -z "$NOT_UPSTREAM" ]; then
	UNIFDEF_DEFINES="-UEFX_NOT_UPSTREAM"
	SOURCE_SED_COMMANDS='s/\bEFX_FATAL\b/netif_err/g; '
	MAKEFILE_SED_COMMANDS='/^ifndef EFX_UPSTREAM/,/^endif # !EFX_UPSTREAM/d; '

	if [ -n "$KCOMPAT" ]; then
		if [ "$AUTOCOMPAT" != "apply" ]; then
			AUTOCOMPAT=keep
		fi
		UNIFDEF_DEFINES="$UNIFDEF_DEFINES -DEFX_USE_KCOMPAT"
	else
		UNIFDEF_DEFINES="$UNIFDEF_DEFINES -UEFX_USE_KCOMPAT"
		MAKEFILE_SED_COMMANDS="$MAKEFILE_SED_COMMANDS"'/^ifndef EFX_NO_KCOMPAT/,/^endif # !EFX_NO_KCOMPAT/d; '
		MAKEFILE_SED_COMMANDS="$MAKEFILE_SED_COMMANDS"'/^#$/d; ' # delete spacing comments
	fi
fi

MAKEFILE_SED_COMMANDS="$MAKEFILE_SED_COMMANDS/^ifdef EFX_NOT_EXPORTED/,/^endif # EFX_NOT_EXPORTED/d; "
MAKEFILE_SED_COMMANDS="$MAKEFILE_SED_COMMANDS/^ifndef NOWERROR/,/^endif # NOWERROR/d;"
UNIFDEF_DEFINES="$UNIFDEF_DEFINES -UEFX_NOT_EXPORTED"

KPATH="$1"
shift

if [ ! -d "$KPATH" ]; then
    echo >&2 "'$KPATH' does not exist"
    exit 2
fi

sources="$*"

if ! UNIFDEF="$(which unifdef 2>/dev/null)"; then
    UNIFDEF=/misc/apps/x86_64/unifdef
    if [ -z "$UNIFDEF" -o ! -e "$UNIFDEF" ]; then
	echo >&2 "unifdef not found; try 'sudo yum install unifdef' or build it from the v5 repository"
	exit 1
    fi
fi

mkdir -p $KPATH/$SUBDIR
rm -f $KPATH/$SUBDIR/*.[ch]

if [ -n "$AUTOCOMPAT" ]; then
    # unifdef can't rewrite an #if unless every single symbol is known.
    # Which means we have to include autocompat even if we're going to
    # apply the symbols anyway
    ./kernel_compat.sh -k $KPATH -q > $KPATH/$SUBDIR/autocompat.h
    if [ "$AUTOCOMPAT" = "apply" ]; then
	# Read in from autocompat.h
	while read prefix word value; do
	    # Need to keep any #undefs after #if defined(EFX_HAVE_XDP_EXT)
	    grep -q "#undef $word" kernel_compat.h && \
		echo ".. keeping $word due to #undef detected in kernel_compat.h" && \
		continue
	    if [  $prefix = "//" ]; then
		UNIFDEF_DEFINES="$UNIFDEF_DEFINES -U$value"
	    elif [ $prefix = "#define" ]; then
		UNIFDEF_DEFINES="$UNIFDEF_DEFINES -D$word"
	    else
		echo >&2 "unable to parse '$prefix $word $value'"
		exit 1
	    fi
	done < $KPATH/$SUBDIR/autocompat.h
    fi
    # Ensure that EFX_USE_KCOMPAT is defined
    SOURCE_SED_COMMANDS="$SOURCE_SED_COMMANDS /\#include \"autocompat.h\"/i\
#define EFX_USE_KCOMPAT"
fi

# Handle Makefiles
# Copy Makefile, deleting unwanted sections
sed "$MAKEFILE_SED_COMMANDS" <Makefile >$KPATH/$SUBDIR/Makefile

TOPDIR=$(echo "$SUBDIR" | sed 's#[a-z]*#..#g')
if [ ! -f $KPATH/Makefile ]; then
    sed "$MAKEFILE_SED_COMMANDS" < $TOPDIR/Makefile > $KPATH/Makefile
fi
if [ -f $TOPDIR/scripts/Makefile.common ]; then
    mkdir -p $KPATH/scripts
    sed "$MAKEFILE_SED_COMMANDS" < $TOPDIR/scripts/Makefile.common > $KPATH/scripts/Makefile.common
fi

# Copy top-level sources, then find required headers and copy them
while [ -n "$sources" ]; do
    missing=""
    for source in $sources; do
	case "$source" in
	    */*.h)
		dest=$KPATH/${SUBDIR:+include}/$source
		;;
	    *)
		dest=$KPATH/$SUBDIR/$source
		;;
	esac
	sourcename=$(basename $source)
	if [ "$sourcename" = kernel_compat.sh -o "$sourcename" = kernel_compat_funcs.sh ]; then
		mkdir -p $(dirname $dest)
		cp $source $dest
		continue
	fi
	if ! [ -f "$source" ]; then
		continue
	fi
	mkdir -p $(dirname $dest)
	if ! [ -f $KPATH/$SUBDIR/$source ]; then
	    if [ -n "$UNIFDEF_DEFINES" ]; then
		# unifdef may return either 0 or 1 for success.
		"$UNIFDEF" $UNIFDEF_DEFINES -k -B <$source \
		    | sed "$SOURCE_SED_COMMANDS" >$dest \
		    && test ${PIPESTATUS[0]} -le 1
	    elif [ -n "$SOURCE_SED_COMMANDS" ]; then
		sed "$SOURCE_SED_COMMANDS" <$source >$dest
	    else
		cp $source $dest
	    fi
	fi
	while read header; do
	    if [ "$header" != config.h -a "$header" != autocompat.h -a \
		 \! -f $KPATH/$SUBDIR/$header ]; then
		missing="$missing $header"
	    fi
	done < <(sed 's/^#include "\([^/]*\)".*$/\1/; t; d' <$dest)
    done
    sources="$missing"
done

# Update .config with our settings
test -f $KPATH/$SUBDIR/Kconfig || exit 0
